#!/usr/bin/env python

"""
Traverse and annotate simplified AST structures.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from simplified import *
import compiler

class Namespace:
    def __init__(self, local_is_structure=0, parent=None):
        if local_is_structure:
            self.local = "structure"
        else:
            self.local = "local"
        self.parent = parent
        self.names = {}
        self.not_local = []

    def make_global(self, name):
        if name not in self.not_local:
            self.not_local.append(name)

    def find(self, name):
        if name not in self.not_local:
            return self.local
        else:
            return "global"

    def store(self, name, types):
        if name not in self.not_local:
            self.names[name] = types
        else:
            self.parent.store(name, types)

    def load(self, name):
        if name in self.not_local or not self.names.has_key(name):
            return self.parent.load(name)
        else:
            return self.names[name]

    def merge(self, namespace):
        for name, types in namespace.names.items():
            if not self.names.has_key(name):
                self.names[name] = types
            else:
                existing = self.names[name]
                for type in types:
                    if type not in existing:
                        existing.append(type)

class System:
    def __init__(self):
        self.count = 0
    def init(self, node):
        if not hasattr(node, "types"):
            node.types = []
    def annotate(self, node, types):
        for type in types:
            if type not in node.types:
                node.types.append(type)
                self.count += 1

class Annotator(Visitor):
    def __init__(self):
        Visitor.__init__(self)
        self.system = System()
        self.types = None
        self.temp = {}

    def process(self, node):
        if hasattr(node, "structure"):
            self.structure = node.structure
            has_structure = 1
        else:
            self.structure = None
            has_structure = 0

        self.namespace = Namespace(self.structure is not None)
        if has_structure:
            node.structure.namespace = self.namespace

        self.visitor = self
        result = self.dispatch(node)

        return result

    def default(self, node):
        for attr in ("args", "params"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        for attr in ("expr", "lvalue", "test", "handler", "star", "dstar"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatch(value))
        for attr in ("body", "else_", "finally_", "code"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        return node

    def dispatch(self, node, *args):
        return Visitor.dispatch(self, node, *args)

    def visitGlobal(self, global_):
        for name in global_.names:
            self.make_global(name)
        return global_

    def visitLoadRef(self, loadref):
        self.types = [loadref.ref]
        return loadref

    def visitLoadName(self, loadname):
        if self.namespace.find(loadname.name) == "structure":
            return self.dispatch(LoadAttr(expr=LoadRef(ref=self.structure), name=loadname.name))
        else:
            self.types = self.namespace.load(loadname, loadname.name)
            return loadname

    def visitStoreName(self, storename):
        if self.namespace.find(storename.name) == "structure":
            return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.structure), name=storename.name, expr=storename.expr))
        else:
            self.namespace.store(storename, storename.name, self.types)
            return storename

    def visitLoadTemp(self, loadtemp):
        index = getattr(loadtemp, "index", None)
        self.types = self.temp[index]
        return loadtemp

    def visitStoreTemp(self, storetemp):
        index = getattr(storetemp, "index", None)
        self.temp[index] = self.types
        return storetemp

    def visitReleaseTemp(self, releasetemp):
        index = getattr(releasetemp, "index", None)
        del self.temp[index]
        return releasetemp

    def visitLoadAttr(self, loadattr):
        self.dispatch(loadattr.expr)
        types = []
        for ref in self.types:
            types += ref.namespace.load(loadattr.name)
        self.types = types
        return loadattr

    def visitStoreAttr(self, storeattr):
        self.dispatch(storeattr.expr)
        expr = self.types
        self.dispatch(storeattr.lvalue)
        for ref in self.types:
            ref.namespace.store(storeattr.name, expr)
        return storeattr

# vim: tabstop=4 expandtab shiftwidth=4
